/*******************************************************************************
 * Copyright (c) 2001 IBM Corporation.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v0.5 
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v05.html
 * 
 * Contributors:
 *     IBM Corporation - example implementation
 ******************************************************************************/

package com.example;

import java.io.IOException;
import java.util.Arrays;
import java.util.Map;

import org.eclipse.core.resources.ICommand;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;

/**
 * An example incremental project builder that copies additional class files
 * from a library package fragment root folder into a Java project's output
 * directory.
 * 
 * General parameters:
 * <ul>
 * <li>The project should be a Java project.</li>
 * <li>The class files are in the "classes" root folder of the project.</li>
 * <li>This builder should run <b>after</b> the Java builder.</li>
 * <li>Full build should copy class files from a secondary library folder
 *   into the output folder maintaining package hierarchy; existing class files
 *   must never be overwritten.</li>
 * <li>Only *.class files should be copied (not other resource files).</li>
 * <li>Incremental build and auto-build should do nothing (except when this
 *    builder has never been invoked before).</li>
 * <li>Changing the project's output folder should be handled.</li>
 * </ul>
 */
public class LibCopyBuilder extends IncrementalProjectBuilder {
	
	/**
	 * Internal debug tracing.
	 */
	static boolean DEBUG = true;
	
	/**
	 * Builder id of this incremental project builder.
	 */	
	public static final String BUILDER_ID = new String(LibCopyPlugIn.ID + ".lcbuilder");

	/**
	 * The path of the output folder that we last copied class files
	 * into, or <code>null</code> if this builder has not built this
	 * project before.
	 */
	private IPath lastOutputPath = null;

	/**
	 * Creates a new instance of the library copying builder.
	 * <p>
	 * All incremental project builders are required to have a public
	 * 0-argument constructor.
	 * </p>
	 */
	public LibCopyBuilder() {
		super();
	}

	/**
	 * Returns the folder containing the additional class files that are
	 * to be copied into the output folder, or <code>null</code> if 
	 * there is none. The contents of the folder are arranged as a standard
	 * package hierarchy (that is, the result is a package fragment root
	 * folder).
	 * 
	 * @return the package fragment root folder, or <code>null</code> if 
	 *    there is none
	 */
	public IFolder getLibraryRootFolder() {
		return getProject().getFolder("classes");
	}

	/**
	 * 
	 * The <code>LibCopyBuilder</code> implementation of this 
	 * <code>IncrementalProjectBuilder</code> method copies additional
	 * class files into the output folder.
	 * <p>
	 * [Issue: the implementation should report progress.]
	 * </p>
	 * <p>
	 * [Issue: the implementation should probably use a workspace runnable.]
	 * </p>
	 */
	protected IProject[] build(int kind, Map args, IProgressMonitor monitor)
		throws CoreException {

		if (DEBUG) {
			System.out.println(BUILDER_ID +
				": Start build project " + getProject().getName());
		}
		
		boolean builderOrderOK = checkBuilderOrdering();

		if (DEBUG && !builderOrderOK) {
			System.out.println(BUILDER_ID + 
				": Bad builder order for project " + getProject().getName());
		}
		
		if (!getLibraryRootFolder().exists()) {
			// no files to copy
			return null;
		}
		
		IJavaProject jproject = JavaCore.create(getProject());
		if (jproject == null) {
			// not a java project (anymore?)
			return null;
		}		
		
		IPath outputPath = jproject.getOutputLocation();
		if (outputPath.equals(lastOutputPath)) {
			if (kind == INCREMENTAL_BUILD || kind == AUTO_BUILD) {
				// we don't usually do anything on incremental or auto-builds
				return null;
			}
		}
		IFolder outputFolder = getProject().getParent().getFolder(outputPath);
		
		copyClassFiles(getLibraryRootFolder(), outputFolder, monitor);
		lastOutputPath = outputPath;

		return null;
	}

	/**
	 * Checks whether this builder is configured to run <b>after</b> the
	 * Java builder.
	 * 
	 * @return <code>true</code> if the builder order is correct, and
	 *    <code>false</code> otherwise
	 * @exception CoreException if something goes wrong
	 */
	private boolean checkBuilderOrdering() throws CoreException {
		// determine relative builder position from project's buildspec
		ICommand[] cs = getProject().getDescription().getBuildSpec();
		int myIndex = -1;
		int javaBuilderIndex = -1;
		for (int i = 0; i < cs.length; i++) {
			if (cs[i].getBuilderName().equals(JavaCore.BUILDER_ID)) {
				javaBuilderIndex = i;
			} else if (cs[i].getBuilderName().equals(BUILDER_ID)) {
				myIndex = i;
			}
		}
		return myIndex > javaBuilderIndex;
	}

	/**
	 * Copies class files from the given source folder to the given 
	 * destination folder. The destination folder will be created if
	 * required, but only if at least one class file is copied.
	 * 
	 * @param source the source folder; must exist
	 * @param dest the destination folder; may or may not exist
	 * @param monitor the progress monitor, or <code>null</code> if none
	 * @exception CoreException if something goes wrong
	 */
	private void copyClassFiles(IFolder source, final IFolder dest,
			final IProgressMonitor monitor) throws CoreException {
				
		if (DEBUG) {
			System.out.println(BUILDER_ID + 
				": Begin copying class files from "
				+ source.getFullPath()
				+ " to "
				+ dest.getFullPath());
		}

		final int sourcePathLength = source.getFullPath().segmentCount();

		class Visitor implements IResourceVisitor {
			public boolean visit(IResource res) throws CoreException {
				if (res.getType() == IResource.FILE) {
					IFile file = (IFile) res;
					String ext = res.getFileExtension();
					if (ext != null && ext.equals("class")) {
						// compute relative path from source folder to this file
						IPath filePath = file.getFullPath();
						IPath dpath = filePath.removeFirstSegments(sourcePathLength);
						IFile targetFile = dest.getFile(dpath);
						copyFile(file, targetFile, monitor);
					}
				}
				return true;
			}
		}

		try {
			source.accept(new Visitor());
		} catch (CoreException e) {
			// should not happen
		}
		
	}

	/**
	 * Copies the given file to the given destination file. Does nothing
	 * if the destination file already exists.
	 * 
	 * @param source the source file; must exist
	 * @param dest the destination file; may or may not exist; never 
	 *    overwritten
	 * @param monitor the progress monitor, or <code>null</code> if none
	 * @exception CoreException if something goes wrong
	 */
	private void copyFile(IFile source, IFile dest, IProgressMonitor monitor) 
			throws CoreException {

		if (dest.exists()) {
			if (DEBUG) {
				System.out.println(BUILDER_ID + 
					": " + dest.getFullPath() + " already exists.");
			}
			return;
		}

		if (DEBUG) {
			System.out.println(BUILDER_ID + 
				": Creating " + dest.getFullPath());
		}

		IContainer parent = dest.getParent();
		if (parent.getType() == IResource.FOLDER) {
			mkdirs((IFolder) parent, monitor);
		}
		dest.create(source.getContents(false), false, monitor);
		
	}

	/**
	 * Creates the given folder, and its containing folders, if required. 
	 * Does nothing if the given folder already exists.
	 * 
	 * @param folder the folder to create
	 * @param monitor the progress monitor, or <code>null</code> if none
	 * @exception CoreException if something goes wrong
	 */
	private void mkdirs(IFolder folder, IProgressMonitor monitor) 
			throws CoreException {
		if (folder.exists()) {
			return;
		}
		IContainer parent = folder.getParent();
		if (!parent.exists() && parent.getType() == IResource.FOLDER) {
			mkdirs((IFolder) parent, monitor);
		}
		folder.create(false, true, monitor);
	}
}

